% Implementation of the SQH mehtod for an optimal tumor therapy problem by
% Melina Kienle and Tim Breitenbach

% Parameters of the control problem
kappa=0.5;
sigma=0.2;
vartheta=1;
mu_u=10000; %L1
mu_w=10000; %L1
nu_u=5000; %L2
nu_w=5000; %L2
T=7;
nu=1;
xi=0.084;
b=5.85;
d=0.00873;
gamma=0.15;
alpha=0.7;
beta=0.14;
eta=0.136;
delta=0.086;
rho=log(2)/0.02;

% Number of discrete time steps
n=1000;
dt=T/n;

% Define struct
OCP=struct('kappa',kappa,'sigma',sigma,'vartheta',vartheta,'mu_u',mu_u,...
           'mu_w',mu_w,'xi',xi,'b',b,'d',d,'gamma',gamma,'alpha',alpha,...
           'beta',beta,'eta',eta,'delta',delta,'rho',rho,'nu_u',nu_u,...
           'nu_w',nu_w,'nu',nu,'T',T,'dt',dt,'n',n);
       
% Initial conditions       
ya=[8000,10000,0];

% Initialize controls
% fid = fopen('u_fT.dat');
% dummy_u = textscan(fid, '%*f %f','delimiter',',','headerlines',1);
% fclose(fid);
% u = dummy_u{1};
% 
% fid = fopen('w_fT.dat');
% dummy_w = textscan(fid, '%*f %f','delimiter',',','headerlines',1);
% fclose(fid);
% w = dummy_w{1};

u=0*ones(n,1);
w=0*ones(n,1);
u_old=u;
w_old=w;

% First calculation of state variable, adjoint variable and objective
[x,t]=expliciteuler(@F,ya,u,w,OCP);
x_old=x;
lambda=eulerback(@H,t,x,u,w,OCP);
J(1)=objective (x(:,1),u,w,OCP)
J_old=J(1);
Jk(1)=J_old;

 
kmax=10000000000; %Maximum iteration number

% Maximum admissible values for the controls
umax=15;
wmax=1;

epsilon=0.1;
count_updates=1; %Counter for the updates of the control
tic
for k=1:kmax
 
       u=min(max(zeros(n,1),(-mu_u*ones(n,1)+lambda(:,2).*gamma.*x(:,2)+2*epsilon.*u)/(nu_u+2*epsilon)),ones(n,1)*umax);
       w=min(max(zeros(n,1),(-mu_w*ones(n,1)+lambda(:,1).*(alpha*ones(n,1)+beta*x(:,3)).*x(:,1)+lambda(:,2).*(ones(n,1)*eta+delta*x(:,3)).*x(:,2)-lambda(:,3)+2*epsilon*w)/(nu_w+2*epsilon)),wmax*ones(n,1));
  
duw=sum((u_old-u).^2+(w_old-w).^2)*dt; %Norm of the control update
   
   [x,t]=expliciteuler(@F,ya,u,w,OCP);
   J_int=objective (x(:,1),u,w,OCP);
   if(J_int-Jk(count_updates)>-(10^-7)*duw)
       u=u_old;
       w=w_old;
       x=x_old;
       epsilon=epsilon*50;                            
   else
       count_updates=count_updates+1;
       lambda=eulerback(@H,t,x,u,w,OCP);
       epsilon=epsilon*0.15;                         
       Jk(count_updates)=J_int;
       u_old=u;
       w_old=w;
       x_old=x;
   end
   if(duw<10^-15)
        k
        u=u_old;
        w=w_old;
        break;
   end 
    
end  
toc
optimal(u,w,ya, umax, wmax, OCP) %Check the PMP optimality of the control

% Print values of objective and optimal states p and q at the terminal time T
sprintf('objective: %f \n p(T): %f \n q(T): %f',Jk(end),x(end,1),x(end,2))

% Plot of optimal states
figure
plot(t,x(:,1),t,x(:,2),'--')
%xlabel('time t in [d]')
axis([0 7.5 0 max(max(x(:,1)),max(x(:,2)))+0.1*max(max(x(:,1)),max(x(:,2)))])
legend('p','q')
title('Optimal states');

% Plot of controls
figure
plot(t,u,t,w,'--')
%xlabel('time t in [d]')
legend('u','w')
axis([0 7.5 0 max(max(u),max(w))+0.1*max(max(u),max(w))])
%ylabel('u,w')
title('Optimal controls');

