%Implementation of the SQH method for an L^0 quantum optimal control problem by Tim Breitenbach 
function [ ] = PMP_QC( )
varkappa=2; %Factor for the weights alpha and beta
%Problem parameters
OCP=struct('dt',10^-5,'T',0.008,'umin',-60,'umax',60,'nu',varkappa*5*10^-3,'beta',varkappa*20);

y0=[0;0;1;0;0;1]; %Initial state
y0=y0/norm(y0);
A=[0,-1,0,0,0,0;1,0,0,0,0,0;0,0,0,0,0,0;0,0,0,0,1,0;0,0,0,-1,0,0;0,0,0,0,0,0]*2*pi*483;
B=[0,0,0,0,0,0;0,0,-1,0,0,0;0,1,0,0,0,0;0,0,0,0,0,0;0,0,0,0,0,-1;0,0,0,0,1,0]*2*pi;

%Tolerance for convergence
kappa=10^-15;                
%Algorithm paramters
eta=10^-9;
zeta=0.8;
sigma=2.0;
kmax=10^10;        %Maximum iteration number
epsilon=0.005;      %Initial guess for epsilon

yd=[0;1;0;0;1;0]; %Desired final state
yd=yd/norm(yd);
Nt=round(OCP.T/OCP.dt);

u=0*ones(1,Nt+1);
u_old=u;

y=forward(A,B,y0,u,OCP);
y_old=y;

p=backward(A,B,y(:,end),yd,u,OCP);

J_plot(1,1)=get_J(y(:,end),u,yd,OCP);
Jk(1)=J_plot(1,1);

count_updates=1;
tic
for k=1:kmax
    for i=1:Nt+1
        %u(1,i)=argminH(y(:,i),u(i),p(:,i),A,B,epsi,OCP ); %Minimizing the augmented Hamiltonian
        
        %Analytical formula for the minimization of the augmented Hamiltonian
        uz(1)=min(max(0,(2*epsilon*u_old(1,i)-(p(:,i)')*B*y(:,i))/(2*epsilon+OCP.nu)),OCP.umax);
        uz(2)=min(max(OCP.umin,(2*epsilon*u_old(1,i)-(p(:,i)')*B*y(:,i))/(2*epsilon+OCP.nu)),0);
        uz(3)=0;
        H(1)=(OCP.nu/2)*uz(1)^2 + OCP.beta*(abs(uz(1))>0) + (p(:,i)')*(A+uz(1)*B)*y(:,i) + epsilon*(uz(1)-u_old(1,i))^2;
        H(2)=(OCP.nu/2)*uz(2)^2 + OCP.beta*(abs(uz(2))>0) + (p(:,i)')*(A+uz(2)*B)*y(:,i) + epsilon*(uz(2)-u_old(1,i))^2;
        H(3)=epsilon*(u_old(1,i)^2);
        [~,pos]=min(H);
        u(1,i)=uz(pos);

    end 
  
   du=sum((u-u_old).^2)*OCP.dt; %Norm of the control update
   
   y=forward(A,B,y0,u,OCP);
   J_int=get_J(y(:,end),u,yd,OCP);
   if(J_int-Jk(count_updates)>-eta*du)
       u=u_old;
       y=y_old;
       epsilon=epsilon*sigma;                            
   else
       count_updates=count_updates+1;
       p=backward(A,B,y(:,end),yd,u,OCP);
       epsilon=epsilon*zeta;                         
       Jk(count_updates)=J_int;
       u_old=u;
       y_old=y;
   end
   if(du<kappa)
        k
        count_updates-1
        u=u_old;
        fprintf('converged\n')
        break;
   end 
end 
toc
figure()
plot(0:OCP.dt/(OCP.T):1,u/100)
title(['$\alpha=$' num2str(OCP.nu) '$\mathrm{,\ }\beta=$' num2str(OCP.beta)],'interpreter','latex')
axis([-inf inf -1 1])
xlabel('t/T')
ylabel('u/100')

figure
plot(Jk)

C=0.5*((y(:,end)-yd)'*(y(:,end)-yd)) %Distance from the state at final time to the desired state

numeric_optimality(u,y0,yd,A,B,OCP) %Check the PMP optimality of the control
end
