%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Implementation of a globalized Newton method by Gabriele Ciaramella
% Main function;
function [ time_vec , u_all , ni_all , functionals , fidelities , coeff_beta , coeff_hatK , bounds36 , normsL1 , maximum , Roots , Roots2 ] = L1SKRYN
%   First-optimize-then-discretize strategy  %

    %	Define the variables;
    global ni
    global N
    global NC
    global T
    global h
    global xd
    global bet
    global theta
    global b
    global lin_tol
    
    % initialize the test;
    [ time_vec , u ] = Test2GL;
    
    % checks from the papier;
    global B
    global x0
    
    % input variables;
    maxit           = 1000; %100
    lin_tol         = 1e-8; %8
    % continuation parameters;
    tol             = 1e-7; %-7
    tolActive       = 1e-5; %-5
    gamma           = 0.5;
    gammaBeta       = 1; %0.5;
    iter_cont       = 0;
    
    % OUTPUT VARIABLES;
    % vectors containing the results from all
    % the continuation procedure;
    fidelities      = zeros(iter_cont,1);
    functionals     = zeros(iter_cont,1);
    u_all           = zeros(iter_cont,N,NC);
    ni_all          = zeros(iter_cont,1);
    coeff_beta      = bet;
    coeff_hatK      = zeros(iter_cont,1);
    bounds36        = zeros(iter_cont,1);
    normsL1         = zeros(iter_cont,1);
    maximum         = zeros(iter_cont,1);
    Roots           = zeros(iter_cont,1);
    Roots2          = zeros(iter_cont,1);
    
    coeff_a         = 0;
    coeff_b         = 0;
    coeff_c         = bet;
    coeff_d         = 0;
    
    % Test of the gradient;
    %GradTest( u );
    
    % initialization;
    numsteps = 1;
    target = xd;
    intermediate = [ 1,1,0  , 1,1,0 ]';
    for k = 1 : numsteps
        fprintf('|--------------------------------------------------------| \n');
        xd = (1-k/numsteps)*intermediate + k/numsteps*target;
        xd = xd/norm(xd);
        tic
        [ u , C , func ] = SKRYN( maxit , tol , tolActive , u , time_vec );
        toc
        plot(time_vec/T,u)
        u=u';
        save('uLONE.mat','u')
        u=u';
        drawnow
        ni = 0.9*ni;
        pause(0.01)
        fprintf('| Initialization Homotopy | step = %4.0f | ni = %12.10e | theta = %12.10e | C = %9.6f | \n', k , ni , theta , C );
        fprintf('|--------------------------------------------------------| \n');
        fprintf('\n');
    end
    xd = target;
    
    % solve the problem with Newton+continuation;
    for cont = 1 : iter_cont
        fprintf('|--------------------------------------------------------| \n');
        %theta = min(min(1/bet,1/ni),1e3);%boundTheta);
        %theta
        tic
        [ u , C , func ] = SKRYN( maxit , tol , tolActive , u , time_vec );
        toc
        figure()
        plot(time_vec/(8*10^-3),u/100)
        axis([-inf inf -1 1])
        drawnow
        pause(0.01)
        % Elaborate datas for output;
        ni_all(cont) = ni;
        u_all(cont,:,:) = u;
        fidelities(cont) = C;
        functionals(cont) = func;
        normsL1(cont) = h * sum( sum( abs(u) ) );
        K0 = 0;
        maximum(cont) = 0;
        for n = 1 : NC
            %K0 = K0 + norm( B(:,:,n) , 'fro' );
            K0 = max( K0 , norm( B(:,:,n) , 'fro' ) );
            maximum(cont) = max( maximum(cont) , (B(:,:,n) * x0)' * xd );
        end
        coeff_hatK(cont) = NC^(1/3) * T^(1/3) / ( 2^(2/3) * norm(x0)^(4/3) * K0^(1/3) );
        bounds36(cont) = 2 * norm(x0)^2*sqrt(2*K0)*min( 1/bet , coeff_hatK(cont)/(ni^(1/3)) );
        % Compute the roots of the third order polinomial;
        coeff_a = ni / ( 2 * T * NC );
        coeff_d = - 2*norm(x0)^2*sqrt(2*K0);
        %Delt = 18*coeff_a*coeff_b*coeff_c*coeff_d-4*coeff_b^3*coeff_d+coeff_b^2*coeff_c^2-4*coeff_a*coeff_c^3-27*coeff_a^2*coeff_d^2;
        Delt0 = coeff_b^2-3*coeff_a*coeff_c;
        Delt1 = 2*(coeff_b^3) - 9*coeff_a*coeff_b*coeff_c + 27*(coeff_a^2)*coeff_d;
        CC = (0.5*( Delt1 + sqrt( Delt1^2 - 4*Delt0^3 ) ))^(1/3);
        Roots(cont) = -(coeff_b+CC+Delt0/CC)/(3*coeff_a);
        vec = roots([coeff_a coeff_b coeff_c coeff_d]);
        Roots2(cont)=vec(3);
        
% %Plot polynomials;
% y_1 = @(x) coeff_d+coeff_c*x+coeff_b*x.^2+coeff_a*x.^3;
% y_lin = @(x) coeff_d+coeff_c*x;
% y_cub = @(x) coeff_d+coeff_b*x.^2+coeff_a*x.^3;
% x_1 = 0:0.01:1;
% plot(x_1,y_1(x_1),'r',x_1,y_lin(x_1),'b',x_1,y_cub(x_1),'m')
% grid on
% ( -coeff_d/coeff_a )^(1/3)
% ( -coeff_d/coeff_c )
% bounds36(cont)
% Roots(cont)
% pause

        % Print;
        fprintf('| Continuation | step = %4.0f | ni = %12.10e | theta = %12.10e | C = %9.6f | \n', cont , ni , theta , C );
        fprintf('|--------------------------------------------------------| \n');
        fprintf('\n');
        ni = gamma * ni;
        bet = gammaBeta * bet;
        if ( C >= 1.0-1e-3 )
            break;
        end
    end
    [ func , grad ,  x , p , U1 , U2 , U3 , U4 ] = FandG( u );
    x(:,N)
    

    % check the subgradient "lambda_bar" and the active set "ActiveS";
    [ func , grad , x , p , U1 , U2 , U3 , U4 , mubar ] = FandG( u );
    lambda_bar = min( bet , max( - bet , mubar ) );
    [ ActiveS ] = ActiveSet( u );
    for n = 1 : NC
        figure(n)
        %plot(time_vec,u(:,n),'bo-',time_vec,lambda_bar(:,n),'ro-',...
        %     time_vec,bet*ones(size(time_vec)),'--k',time_vec,-bet*ones(size(time_vec)),'--k',...
        %     'LineWidth',1.4);
%         plot(time_vec(2:end-1),u(2:end-1,n),'-k',time_vec(2:end-1),lambda_bar(2:end-1,n),'--k',...
%              time_vec,bet*ones(size(time_vec)),':k',time_vec,-bet*ones(size(time_vec)),':k',...
%              'LineWidth',2);
%         string1 = ['u_{',num2str(n),'}(t)'];
%         string2 = ['\lambda_{',num2str(n),'}(t)'];
%         legend(string1,string2,'\beta','-\beta','Location','SouthEast' );
        %axis([ 0 time_vec(end)   -1 1 ]);
        %if(b<1e6)
        %	axis([ 0 time_vec(end)   -(max(b,bet)+0.05) (max(b,bet)+0.05) ]);
        %else
        %	axis([ 0 time_vec(end)   -(bet+0.05) (bet+0.05) ]);
        %end
%        grid on;
        %figure(n+NC)
        %plot(time_vec,abs(u(:,n)),'bo-',time_vec,b*ActiveS(:,n),'ro-',...
        %     time_vec,b*ones(size(time_vec)),'k',time_vec,-b*ones(size(time_vec)),'k',...
        %     'LineWidth',1.4);
        %string1 = ['|u_{',num2str(n),'}(t)|'];
        %string2 = ['b*A_{',num2str(n),'}(t)'];
        %legend(string1,string2,'b','Location','NorthEast' );
        %axis([ 0 time_vec(end)   0 (b+0.05) ]);
        %grid on;
    end

end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SKRYN: semismooth matrix-free Krylov-Newton method;
function [ u , C , func ] = SKRYN( maxit , tol , tolActive , u , time_vec )
    % global variables;
    global N
    global NC
    global h
    global xd
    global bet
    global ni
    global lin_tol
    
    fprintf('Krylov-NEWTON method\n');

    % initialize the method;
    [ func , grad , x , p , U1 , U2 , U3 , U4 ] = FandG( u );  
    nor = sqrt( h * sum( grad.^2 ) );
    [ ActiveS ] = ActiveSet( u );
    ActiveS_old = 0;
    numAS = sum(sum(ActiveS));
    du = zeros(size(u));

    % iterations;
    alpha = 0;
    k = 0;
    direct = -1e1;
    linesearch = '------';
    counter = 0;
    for it = 1 : maxit
        
        % print;
        C = sum( x(:,N) .* xd ) / (norm(xd)^2);
        fprintf('| it = %4.0f | func = %14.8f | ActiveS = %4.0f | grad = %3.8e | direct = %3.8e | lineS = %6s | alpha = %14.8f | k = %4.0f | C = %9.6f |\n',...
                 it-1 , func , numAS , nor , direct , linesearch , alpha , k , C );
        
        % stopping criterion;
        if ( sum(sum(ActiveS))==0 )
            if( nor <= tol )
                fprintf('Stop: residual!\n');
                break;
            end
        else
            dif = sum(sum(ActiveS-ActiveS_old));
            if ( dif == 0 && nor <= tolActive )
                fprintf('Stop: Active residual!\n')
                break;
            end
            if( abs(direct) <= tolActive*1e-3 )
                fprintf('Stop: Active direction!\n')
                break;
            end
        end
        if( abs(direct) <= tol*1e-3 )
            fprintf('Stop: direction!\n')
            break;
        end
        
        % solve the Newton linear system;
        duV = - grad / ni;
        [ duV , flag ] = gmres( @( duV ) ApplyH ( duV , u , x , p , U1 , U2 , U3 , U4 ) , -grad , [] , lin_tol , N*NC , [] , [] , duV );
        %[ duV , flag ] = minres( @( duV ) ApplyH ( duV , u , x , p , U1 , U2 , U3 , U4 ) , -grad , lin_tol , 20 );
        %[ duV , flag ] = bicgstab( @( duV ) ApplyH ( duV , u , x , p , U1 , U2 , U3 , U4 ) , -grad , lin_tol , N*NC , [] , [] , duV );
        
        if( flag > 0 )
            fprintf('Warning: flag = %4.0f \n', flag );
            if ( flag > 1 )
                [ duV , flag ] = bicgstab( @( duV ) ApplyH ( duV , u , x , p , U1 , U2 , U3 , U4 ) , -grad , lin_tol , 20 );
                %[ duV , flag ] = minres( @( duV ) ApplyH ( duV , u , x , p , U1 , U2 , U3 , U4 ) , -grad , lin_tol , 20 );
            end
        end

        % check the direction;
        direct = h * sum( grad .* duV );
        if ( numAS == 0 && direct > 0 )
            fprintf('Warning: direction = %16.8f \n', direct );
            duV = - duV;
        end
        
        for i = 1 : NC
            du(:,i) = duV((i-1)*N+1:i*N);
        end
        
        % line-search;
        if ( numAS == 0 )
            if ( bet == 0 )
                linesearch = 'beta=0';
                [ alpha , k ] = lineS( func , u , grad , du , duV );
                %[ alpha , info ] = linesearchSWCBisec ( func , grad , duV , u );
                %k = 999;
            else
                linesearch = 'beta>0';
                [ alpha , k ] = NoDerivNonSmoothLineSearch( func , du , u , it );
            end
        else
            %alpha = 1.0;
            %[ alpha , k ] = semismoothLineSearch( du , u );
            linesearch = 'active';
            [ alpha , k ] = NoDerivLineSearch( du , u , it );
            %[ alpha , k ] = NoDerivLineSearchBisec( du , u , it );
        end
        
        % update unknown;
        u = u + alpha * du;
        if ( alpha < 1e-6 )
            fprintf(' Warning alpha too small!! \n');
            if ( counter == 1 )
                fprintf(' Warning alpha too small!! \n');
                break;
            end
            %alpha = 1.0;
            counter = 1;
        end

        % evaluate the gradient;
        [ func , grad ,  x , p , U1 , U2 , U3 , U4 ] = FandG( u );
        grad = reshape(grad,N*NC,1);
        
        % norm of the gradient;
        nor = sqrt( h * sum( grad.^2 ) );

	% construct the active set;
        ActiveS_old = ActiveS;
    	[ ActiveS ] = ActiveSet( u );
    	numAS = sum(sum(ActiveS));
        
    end
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% lineS: line-search Armijio's rule;
function [ alpha , k ] = lineS( func , u , grad , du , duV )
    
    % global variables;
    global h

    % line-search;
    gamm = 1e-4;
    %gamma = 1e-6;
    alpha = 1.0;
    nor1 = h * sum( grad.*duV );
    for k = 1 : 20
        % recompute the cost functional;
        [ x2 ] = constr( u+alpha*du );
        %[ p2 ] = adjoint( u+alpha*du , x2 );
        [ func2 ] = functional( u+alpha*du , x2 );
        % Armijio's condition;
        if( func2 <= func + alpha * gamm * nor1 )
           break;
        end
        % update rho;
        alpha = alpha * 0.5;
    end
    
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% NoDerivLineSearch: line-search for the "active set case" (it is an Armijio-type without derivatives!);
function [ alpha , k ] = NoDerivLineSearch( du , u , it )
    global h
    
    gamm = 1e-6;
    kmax = 20;
    delta = 0.5;
    rho = 1e-3^it;
    vari = h*sum(sum(du.^2));

    % choose an initial "alpha";
    %a = 1e4;
    alpha = 1.0;
    %if( alpha < a / sqrt(vari) )
    %    fprintf(' Warning No-derivative line-search: alpha<a||du|| \n');
    %    alpha = a / sqrt(vari);
    %end
    
    [ func , grad , x , p , U1 , U2 , U3 , U4 ] = FandG( u );
    theta = 0.5 * h * sum( grad.^2 );
    
    for k = 1 : kmax
        %if( alpha * sqrt(vari) < rho )
        %    alpha = 0;
        %    fprintf(' Warning No-derivative line-search: alpha*||du||<rho \n');
        %    break;
        %end
        u_test = u + alpha * du;
        [ func , grad_test , x , p , U1 , U2 , U3 , U4 ] = FandG( u_test );
        theta_test = 0.5 * h * sum( grad_test.^2 );
        if( theta_test <= theta - gamm * alpha^2 * vari )
            break;
	else
	    u_test = u - alpha * du;
	    [ func , grad_test , x , p , U1 , U2 , U3 , U4 ] = FandG( u_test );
            theta_test = 0.5 * h * sum( grad_test.^2 );
            if( theta_test <= theta - gamm * alpha^2 * vari )
                alpha = - alpha;
                break;
            end
        end
        alpha = alpha * delta;
    end

    if( k >= kmax )
        fprintf(' Warning No-derivative line-search: reached kmax!! \n');
    end

end
%%%%%%%%%%%%%%%%%%%
function [ alpha , k ] = NoDerivLineSearchBisec( du , u , it )
    global h
    
    gamm = 1e-6;
    kmax = 20;
    vari = h*sum(sum(du.^2));

    % choose an initial "alpha";
    alpha = 1.0;
    
    [ func , grad , x , p , U1 , U2 , U3 , U4 ] = FandG( u );
    theta0 = 0.5 * h * sum( grad.^2 );
    [ func , grad , x , p , U1 , U2 , U3 , U4 ] = FandG( u+alpha*du );
    thetaMax = 0.5 * h * sum( grad.^2 );
    if( thetaMax <= theta0 - gamm * alpha^2 * vari )
            k = 0;
            return;
    end
            
    for k = 1 : kmax
        u_test = u + alpha/2 * du;
        [ func , grad_test , x , p , U1 , U2 , U3 , U4 ] = FandG( u_test );
        theta_test = 0.5 * h * sum( grad_test.^2 );
        if( theta_test <= theta0 - gamm * (alpha/2)^2 * vari )
            alpha = alpha/2;
            return;
        end
        if( theta_test <= thetaMax && theta_test <= theta0 )
            if( theta0 <= thetaMax )
                alpha = alpha/2;
                [ func , grad , x , p , U1 , U2 , U3 , U4 ] = FandG( u+alpha*du );
                thetaMax = 0.5 * h * sum( grad.^2 );
            else
                alpha0 = alpha/2;
                [ func , grad , x , p , U1 , U2 , U3 , U4 ] = FandG( u+alpha0*du );
                theta0 = 0.5 * h * sum( grad.^2 );
            end
        else
            if( theta_test <= thetaMax )
                alpha = alpha/2;
                [ func , grad , x , p , U1 , U2 , U3 , U4 ] = FandG( u+alpha*du );
                thetaMax = 0.5 * h * sum( grad.^2 );
            else
                alpha0 = alpha/2;
                [ func , grad , x , p , U1 , U2 , U3 , U4 ] = FandG( u+alpha0*du );
                theta0 = 0.5 * h * sum( grad.^2 );
            end
        end
    end

    if( k >= kmax )
        fprintf(' Warning No-derivative line-search: reached kmax!! \n');
    end

end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% NoDerivLineSearch: line-search for the "active set case" (it is an Armijio-type without derivatives!);
function [ alpha , k ] = NoDerivNonSmoothLineSearch( func , du , u , it )
    global h
    
    gamm = 1e-6;
    kmax = 20;
    delta = 0.5;
    rho = 1e-3^it;
    vari = h*sum(sum(du.^2));

    % choose an initial "alpha";
    %a = 1e4;
    alpha = 1.0;
    %if( alpha < a / sqrt(vari) )
    %    fprintf(' Warning No-derivative line-search: alpha<a||du|| \n');
    %    alpha = a / sqrt(vari);
    %end
    
    %while ( delta < 1 )
        for k = 1 : kmax
            %if( alpha * sqrt(vari) < rho )
            %    alpha = 0;
            %    fprintf(' Warning No-derivative line-search: alpha*||du||<rho \n');
            %    break;
            %end
            u_test = u + alpha * du;
            [ func_test ] = FandG( u_test );
            if( func_test <= func - gamm * alpha^2 * vari )
                return;
        else
            u_test = u - alpha * du;
            [ func_test ] = FandG( u_test );
                if( func_test <= func - gamm * alpha^2 * vari )
                    alpha = - alpha;
                    break;
                end
            end
            alpha = alpha * delta;
        end

        if( k >= kmax )
            fprintf(' Warning No-derivative line-search: reached kmax!! \n');
            %gamm = 0;
            %delta = delta + 0.2;
            %if (delta < 1)
            %    alpha = 1.0;
            %end
            %kmax = 3*kmax;
        end
    %end

end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% FandG: compute the cost functional and the gradient;
function [ func , grad , x , p , U1 , U2 , U3 , U4 , mubar ] = FandG( u )

    % Solve the constraint "x" forward;
    [ x ] = constr( u );
    
    % Compute the cost functional;
    [ func ] = functional( u , x );
    
    if ( nargout > 1 )
	    % Solve the constraint "p" backward;
	    [ p ] = adjoint( u , x );
        if ( nargout < 9 )
            % Assemble the gradient;
                [ grad , U1 , U2 , U3 , U4 ] = gradient( u , x , p );
        else
            [ grad , U1 , U2 , U3 , U4 , mubar ] = gradient( u , x , p );
        end
    end
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ApplyH: apply the reduced generalized Jacobian on a vector function du;
function [ Appl ] = ApplyH( duV , u , x , p , U1 , U2 , U3 , U4 )

    global N
    global NC
    
    du = zeros(size(u));
    for i = 1 : NC
        du(:,i) = duV((i-1)*N+1:i*N);
    end

    % Solve the linearized constraint "dx" forward;
    [ dx ] = lin_constr( du , u , x );
    % Solve the linearized constraint "dp" backward;
    [ dp ] = lin_adjoint( du , dx , u , p );
    
    % Assemble the action;
    [ Appl ] = lin_gradient( duV , dx , dp , u , x , p , U1 , U2 , U3 , U4 );

end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ActiveSet: function constructing the active set of the controls;
function [ ActiveS ] = ActiveSet( u )
    global N
    global NC
    global b
    indeces = find( abs(u)+1e-8>=b );
    ActiveS = zeros(size(u));
    for k = 1 : size(indeces,1)
	for j = 1 : size(indeces,2)
		ActiveS(indeces(k,j)) = 1;
	end
    end
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% GradTest: Test of the gradient;
function GradTest( u )
    fprintf('Test of the discrete gradient and discrete Hessian \n');
    global h
    global N
    global NC

    % Test gradient;
    u=2*rand(size(u))-1;
    u(1:10,:) = 0*u(1:10,:);
    
    du=rand(size(u));
    du(1,:) = 0*du(1,:);
    
    %solve the constraint;
    [ x ] = constr( u ); 
    % solve the adjoint;
    [ p ] = adjoint( u , x );
    % compute the cost functional;
    [ func ] = functional( u , x );
    % compute the gradient;
    [ grad , U1 , U2 , U3 , U4 ] = gradient( u , x , p );
    
    % compute the action of the Hessian;
    duV = reshape(du,N*NC,1);
    [ Appl ] = ApplyH( duV , u , x , p , U1 , U2 , U3 , U4 );

    epsi = 1.0;
    for j = 1 : 6
      [ xP ] = constr( u+epsi*du );
      [ pP ] = adjoint( u+epsi*du , xP );
      [ funcP ] = functional( u+epsi*du  , xP );
      [ gradP ] = gradient( u+epsi*du , xP , pP );
      
      [ xM ] = constr( u-epsi*du );
      [ pM ] = adjoint( u-epsi*du , xM );
      [ funcM ] = functional( u-epsi*du  , xM );
      [ gradM ] = gradient( u-epsi*du , xM , pM );
      
      left = (funcP-funcM)/(2*epsi);
      %left = (funcP-func)/epsi;
      %left = (func-funcM)/epsi;
      right = h * sum( grad.*duV );
      test = abs(left - right);
      
      %testH = 0;
      testH = h * sum( ( ( gradP - gradM ) / ( 2 * epsi ) - Appl ).^2 );
      testH = sqrt( testH );
      
      fprintf('epsi %e left %e right %e test %e testH %e\n', epsi , left , right , test , testH );
      epsi = 0.1*epsi;
    end
    pause
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Problem functions;
% cost functional
function [ func ] = functional( u , x )
    global N
    global h
    global xd
    global ni
    global bet
    func = 0.5 * sum( (x(:,N) - xd).^2 )...
           + 0.5 * h * ni * sum( sum( u(2:N,:).^2 ) )...
           + h * bet * sum( sum( abs(u(2:N,:)) ) );
end

% constraint integrator;
function [ x ] = constr( u )
    global N
    global NC
    global h
    global x0
    global A B
    
    x = zeros( size(x0,1) , N );
    Id = eye(size(x0,1));
    x(:,1) = x0;
    for k = 2 : N
        M = zeros(size(x0,1));
        for n = 1 : NC
            M = M + (u(k,n)+u(k-1,n)) * B(:,:,n);
        end
        P = ( Id - 0.25 * h * ( 2 * A + M ) ) \ ( Id + 0.25 * h * ( 2 * A + M ) );
        x(:,k) = P * x(:,k-1);
    end
end

% adjoint integrator;
function [ p ] = adjoint( u , x )
    global N
    global NC
    global h
    global x0 xd
    global A B
    global ni
    
    p = zeros( size(x0,1) , N );
    M2 = zeros(size(x0,1));
    Id = eye(size(x0,1));
    for n = 1 : NC
        M2 = M2 + (u(N-1,n)+u(N,n)) * B(:,:,n);
    end
    p(:,N) = ( Id - 0.25 * h * ( 2 * A + M2 )' ) \ ( - ( x(:,N) - xd ) );
    
    for k = N-1 : -1 : 2
        M1 = zeros(size(x0,1));
        for n = 1 : NC
            M1 = M1 + (u(k,n)+u(k-1,n)) * B(:,:,n);
        end
        p(:,k) =  ( Id - 0.25 * h * ( 2 * A + M1 )' ) \ ( ( Id + 0.25 * h * ( 2 * A + M2 )' ) * p(:,k+1) );
	M2 = M1;
    end
end

% gradient assembler;
function [ grad , U1 , U2 , U3 , U4 , mubar ] = gradient( u , x , p )
    global N
    global NC
    global h
    global A B
    global ni
    global b
    global bet
    global theta    

    % construct mubar (switching function);
    gradL2 = zeros( size(u) );
    mubar = zeros( size(u) );
    for k = 2 : N-1
        for n = 1 : NC
            gradL2(k,n) = ni * u(k,n) - 0.25 * ( ( B(:,:,n) * ( x(:,k) + x(:,k+1) ) )' * p(:,k+1) ) - 0.25 * ( ( B(:,:,n) * ( x(:,k) + x(:,k-1) ) )' * p(:,k) );
        end
    end
    for n = 1 : NC
        gradL2(N,n) = ni * u(N,n) - 0.25 * ( ( B(:,:,n) * ( x(:,N) + x(:,N-1) ) )' * p(:,N) );
    end
    mubar = - gradL2;

    % construct the first-order opt. condition (omplementarity condition);
    % % complementarity condition in the form of Stadler (formula (2.10)!).
    %grad2 = u - max( 0 , u + theta * ( mubar - bet ) )...
    %         - min( 0 , u + theta * ( mubar + bet ) )...
    %         + max( 0 , ( u - b ) + theta * ( mubar - bet ) )...
    %         + min( 0 , ( u + b ) + theta * ( mubar + bet ) );
    U1 = zeros(size(mubar));
    U2 = zeros(size(mubar));
    U3 = zeros(size(mubar));
    U4 = zeros(size(mubar));	
    for j = 1 : size(mubar,2)
    	for i = 1 : size(mubar,1)
                if( u(i,j) + theta * ( mubar(i,j) - bet ) >= 0 )
    			U1(i,j) = 1 ;
    		end
    		if( u(i,j) + theta * ( mubar(i,j) + bet ) <= 0 )
    			U2(i,j) = 1 ;
    		end
		if( u(i,j) - b + theta * ( mubar(i,j) - bet ) >= 0 )
    			U3(i,j) = 1 ;
    		end
    		if( u(i,j) + b + theta * ( mubar(i,j) + bet ) <= 0 )
    			U4(i,j) = 1 ;
    		end
    	end
    end

    grad = u - (u+theta*(mubar-bet)).*U1...
             - (u+theta*(mubar+bet)).*U2...
             + (u-b+theta*(mubar-bet)).*U3...
             + (u+b+theta*(mubar+bet)).*U4;

    %plot(grad,'bo',grad2,'r')
    %pause
    grad = reshape(grad,N*NC,1);

end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Linearized equation functions;
% linearized constraint integrator;
function [ dx ] = lin_constr( du , u , x )
    global N
    global NC
    global h
    global x0
    global A B
    
    dx = zeros( size(x0,1) , N );
    Id = eye(size(x0,1));
    dx(:,1) = 0*x0;
    for k = 2 : N
        M1 = zeros(size(x0,1));
        dM1 = zeros(size(x0,1));
        for n = 1 : NC
            dM1 = dM1 + (du(k,n)+du(k-1,n)) * B(:,:,n);
            M1 = M1 + (u(k,n)+u(k-1,n)) * B(:,:,n);
        end
        dx(:,k) = ( Id - 0.25 * h * ( 2 * A + M1 ) ) \ ( ( Id + 0.25 * h * ( 2 * A + M1 ) ) * dx(:,k-1) + 0.25 * h * dM1 * ( x(:,k) + x(:,k-1) ) );
    end 
end

% linearized adjoint integrator;
function [ dp ] = lin_adjoint( du , dx , u , p )
    global N
    global NC
    global h
    global x0 xd
    global A B
    
    dp = zeros( size(x0,1) , N );
    M2 = zeros(size(x0,1));
    dM2 = zeros(size(x0,1));
    Id = eye(size(x0,1));
    for n = 1 : NC
        M2 = M2 + (u(N-1,n)+u(N,n)) * B(:,:,n);
	dM2 = dM2 + (du(N-1,n)+du(N,n)) * B(:,:,n);
    end
    dp(:,N) = ( Id - 0.25 * h * ( 2 * A + M2 )' ) \ ( - dx(:,N) + 0.25 * h * dM2' * p(:,N) );
    
    for k = N-1 : -1 : 2
        M1 = zeros(size(x0,1));
        dM1 = zeros(size(x0,1));
        for n = 1 : NC
            M1 = M1 + (u(k,n)+u(k-1,n)) * B(:,:,n);
            dM1 = dM1 + (du(k,n)+du(k-1,n)) * B(:,:,n);
        end
        dp(:,k) =  ( Id - 0.25 * h * ( 2 * A + M1 )' ) \ ( ( Id + 0.25 * h * ( 2 * A + M2 )' ) * dp(:,k+1) + 0.25 * h * dM2' * p(:,k+1) + 0.25 * h * dM1' * p(:,k) );
	M2 = M1;
	dM2 = dM1;
    end
end

% linearized gradient assembler;
function [ Action ] = lin_gradient( duV , dx , dp , u , x , p , U1 , U2 , U3 , U4 );
    global N
    global NC
    global h
    global A B
    global ni
    global theta

    du = zeros(size(u));
    for i = 1 : NC
        du(:,i) = duV((i-1)*N+1:i*N);
    end
    
    % construction of the variation "dmubar";
    dmubar = zeros( size(u) );
    for k = 2 : N-1
        for n = 1 : NC
            dmubar(k,n) = - ni * du(k,n)...
                    + 0.25 * ( ( B(:,:,n) * ( dx(:,k) + dx(:,k+1) ) )' * p(:,k+1) ) + 0.25 * ( ( B(:,:,n) * ( dx(:,k) + dx(:,k-1) ) )' * p(:,k) )...
                    + 0.25 * ( ( B(:,:,n) * ( x(:,k) + x(:,k+1) ) )' * dp(:,k+1) ) + 0.25 * ( ( B(:,:,n) * ( x(:,k) + x(:,k-1) ) )' * dp(:,k) );
        end
    end
    for n = 1 : NC
        dmubar(N,n) = - ni * du(N,n)...
                + 0.25 * ( ( B(:,:,n) * ( dx(:,N) + dx(:,N-1) ) )' * p(:,N) )...
                + 0.25 * ( ( B(:,:,n) * ( x(:,N) + x(:,N-1) ) )' * dp(:,N) );
    end
    
    % assemble the action;
    %Action2 = - dmubar;
    Action = zeros( size(u) );
    Action = du - (du+theta*dmubar).*U1...
                - (du+theta*dmubar).*U2...
                + (du+theta*dmubar).*U3...
                + (du+theta*dmubar).*U4;
    %plot(Action,'bo',Action2,'r')
    %pause
    Action = reshape(Action,N*NC,1);
    
    %dmubar-ni*du
    %norm(dmubar)
    %norm(Action)
    %pause

end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [ alpha , info ] = linesearchSWCBisec ( func , grad , d , u )

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   STEP 0-a: Declaration of the used global variables;                                 %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    global h
    global ni
    global N
    global NC

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   STEP 0-b: Definition of the Armijo-Wolfe coefficients;                              %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    c1 = 0.10;
    c2 = 0.90;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   STEP 0-c: Compute the variables corresponding to "alpha=0";                         %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    var1 = h * sum( grad .* d );
    var = var1;
    var2 = var1 * c2;
    var1 = var1 * c1;
    
    uV = reshape(u,N*NC,1);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   STEP 1: Choose the linesearch strategy;                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    m2 = 0.5 * ni * h * sum( d.^2 );

    if ( m2 > 1e-14 ) 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   STEP 1-a: STRONG STRATEGY: Winckel-Borzi-Volkwein strategy (see the paper...)       %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        info = 'Strong';
         % estimate "alpha_max" by means of the quadratic model "m2*alpha^2+m1*alpha+m0<=J(u+alpha*d)";
         m0 = h * sum( uV.^2 );
         m0 = 0.5 * ni * h * sum( uV.^2 ) - func;
         m1 = ni * h * sum( uV .* d );
         alpha_max = ( sqrt( m1^2 - 4.0 * m0 * m2 ) - m1 ) / ( 2.0 * m2 );
         u_test = u + alpha_max * d;
         %[ func_max , ~ , grad_test , ~ , ~ , ~ , ~ ] = fandg ( u_test );
         [ func_max , grad_test ] = FandG( u_test );
         var3_max = h * sum( d .* grad_test );

         if ( alpha_max > 2.0 )
            %check the SWC with "alpha=1";
            alpha = 1.00;
            %alpha = alpha_max;
            u_test = u + alpha * d;
            %[ func_1 , ~ , grad_test , ~ , ~ , ~ , ~ ] = fandg ( u_test );
            [ func_1 , grad_test ] = FandG( u_test );
            %SWC Condition
            var3 = func + alpha * var1;
            var3_1 = h * sum( d .* grad_test );
            if ( func_1 <= var3 && abs(var3_1) <= abs(var2) ) %Armijo & Wolfe
                    return;
            else
                %compute alpha_new by means of cubic model on [0,1];
                q1 = ( 3.0 * ( func_1 - func ) - var3_1 - 2.0 * var );
                q2 = ( - 2.0 * ( func_1 - func ) + var3_1 + var );
                alpha_m = var / ( - q1 - sqrt( q1^2 - 3.0 * var * q2 ) );
                alpha_m = real(alpha_m);

                %check the SWC with "alpha_m";
                u_test = u + alpha_m * d;
                %[ func_m , ~ , grad_test , ~ , ~ , ~ , ~ ] = fandg ( u_test );
                [ func_m , grad_test ] = FandG( u_test );
                var_m = func + alpha_m * var1;
                var3_m = h * sum( d .* grad_test );
                if ( func_m <= var_m && abs(var3_m) <= abs(var2) ) %Armijo & Wolfe
                        alpha = alpha_m;
                        return;
                else
                    %need to perform a bisection: find the "right" extremum of the interval;
                    %is the interval [0,alpha_m] ?
                    if ( var1 < 0.0 && ( var3_m > 0.0 || func_m > func ) || var1 > 0.0 && func_m < func )
                        alpha_r = alpha_m;
                        func_r = func_m;
                        var3_r = var3_m;
                    else
                        %is the interval [0,1] ?
                        if ( var1 < 0.0 && ( var3_1 > 0.0 || func_1 > func ) || var1 > 0.0 && func_1 < func )
                            alpha_r = 1.0;
                            func_r = func_1;
                            var3_r = var3_1;
                        else
                            %assume "alpha_r=alpha_max", so the interval is [0,alpha_max];
                            alpha_r = alpha_max;
                            func_r = func_max;
                            var3_r = var3_max;
                        end
                    end
                    %bisection in [0,alpha_r];
                    [ alpha ] = bisection ( 0.0 , func , var , alpha_r , func_r , var3_r , u , d , var1 , var2 );
                end
            end
         else
            %bisection in [0,alpha_max];
            [ alpha ] = bisection ( 0.0 , func , var , alpha_max , func_max , var3_max , u , d , var1 , var2 );
         end

    else
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   STEP 1-b: WEAK STRATEGY: Armijio's rule;                                            %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        info = 'Weak';
        %check the Armijio's condition with "alpha=1";
        alpha = 1.00;
        u_test = u + alpha * d;
        [ func_1 ] = FandG ( u_test );
        %Armijio's Condition
        var3 = func + alpha * var1;
        if ( func_1 <= var3 ) %Armijo
                return;
        else
            % backtracking...
            k = 0;
            kmax = 100;
            while ( k <= kmax )
                %evaluate the guess value u_new;
                u_new = u + alpha .* d;
                %compute the functional;
                [ func_new ] = FandG ( u_new );
                %Armijo Condition
                var3 = func + alpha * var1;
                if ( func_new <= var3 ) %Armijo
                    break;
                else
                    alpha = alpha * 0.9;
                end
                % update k;
                k = k + 1;
            end
        end

    end

end

function [ alpha ] = bisection ( alpha_l , f_l , d_l , alpha_r , f_r , d_r , u , d , var1 , var2 )

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   STEP 0: Declaration of the used the global variables;                               %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    global h

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   STEP 1: Initialization of the procedure;                                            %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    tol = 1e-5;
    L = alpha_r - alpha_l;
    f = f_l;
    alpha = alpha_l;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   STEP 2: Perform the bisection over the given interval;                              %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    while ( L > tol )
        alpha = 0.5 * ( alpha_r + alpha_l );
        u_a = u + alpha * d;
        %[ f_a , ~ , grad_a , ~ , ~ , ~ , ~ ] = fandg ( u_a );
        [ f_a , grad_a ] = FandG( u_a );
        d_a = h * sum( d .* grad_a );
        %SWC Condition
        var = f + alpha * var1;
        if ( f_a <= var ) %Armijo
            var = h * sum( d .* grad_a );
            if( abs(var) <= abs(var2) ) %Wolfe
                return;
            end
        end
        if ( d_l < 0.0 && ( d_r > 0 ||  f_r > f_l ) )
            alpha_r = alpha;
            f_r = f_a;
            d_r = d_a;
        else
            if ( d_l > 0.0 && ( d_r < 0 ||  f_r < f_l ) )
                alpha_r = alpha;
                f_r = f_a;
                d_r = d_a;
            else
                alpha_l = alpha;
                f_l = f_a;
                d_l = d_a;
            end
        end
        L = alpha_r - alpha_l;
    end

end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
