%% Chang-Cooper matrix
% Calculate the matrix of the Chang-Cooper discretization of the
% differential operator

function [M] = matrix_cc(NX, dx, x, dt, t, u)

%% Initialization
th = 70; % threshold: when w is too big, its exp explodes
w =  dx .* B(x,u) ./ C(x,t,u) ;
delta_cc = w .* 0; % initialization delta chang chooper
id_w_big = ( w >  th ); % indexes where w is very big, and its exp would explode
id_w_big_= ( w < -th ); % indexes where exp(w) is very close to -infty

id_w_0 = (abs(w) < 1e-10); % indexes where w is close to zero   

delta_cc(id_w_0) = 0.5;  % where w is null, delta = 1/2
delta_cc(id_w_big) = 0;
delta_cc(id_w_big_)= 1;

id_w_mid = ~(id_w_big_ | id_w_0 | id_w_big);

delta_cc( id_w_mid ) = 1 ./ w(id_w_mid) - 1 ./ (exp(w(id_w_mid)) - 1);
% where w is far from zero, delta can be computed

%% Matrix Chang-Cooper
% Mian diagonal. B_1 and B_end include the no-flux boundary conditions

Btilda = - ( (  C( x(1:end)+dx/2, u )  +  C( x(1:end)-dx/2, u )  )/dx -...
             delta_cc(1:end) .* B(x(1:end)+dx/2,u) ...
             + ( 1 - delta_cc(1:end)) .* B(x(1:end)-dx/2,u)...
             ) / dx;


% Diagonal +1   
Atilda = (1/dx)*( B( x(1:end-1) + dx/2, u ).*(1-delta_cc(1:end-1)) + C( x(1:end-1) + dx/2, u )/dx ) ;

% Diagonal -1
Ctilda = -( B( x(2:end) - dx/2, u ) .* delta_cc(2:end) - ...
           C( x(2:end) - dx/2, u ) / dx ) / dx;

% Assembling
M = diag(Btilda) + diag(Atilda, +1) + diag(Ctilda,-1);

% 
%     w = dx .* B(x+dx/2,u) ./ C(x+dx/2,t_old,u); 
%     
%     if norm(w) < 1.0e-10
%         W = ones(1, NX+1);
%     else
%         W = w ./ (exp(w)-1); 
%     end
%     % diag +1
%     Atilda = C(x(1:end-1) + dx/2,t_old,u) .* W(1:end-1) .* exp(w(1:end-1)) ./ (dx^2);
%     % main diag
%     Btilda = (C(x(2:end-1)+dx/2,t_old,u) .* W(2:end-1) + ...
%                     C(x(2:end-1)-dx/2,t_old,u) .* W(2:end-1) .* exp(w(2:end-1))) ./ (dx^2);
%     Btilda = [C(x(1)+dx/2,t_old,u) * W(1) / (dx^2),... % BC in x_1
%               Btilda,...
%               (C(x(end)-dx/2,t_old,u) * W(end-1) * exp(w(end-1))) / (dx^2)]; % BC in x_N
%     % diag -1
%     Ctilda = (C(x(2:end)-dx/2,t_old,u) .* W(1:end-1) ) ./ (dx^2);
%     
%     M = diag(-Btilda) + diag(Atilda, +1) + diag(Ctilda,-1);

end