function [f_num] = solve_FP(dx, x, NX, NT, dt, t, f_iniz, G, u)

f_num = zeros(NX + 1, NT + 1);  
% Initial condition
f_num(:,1) = f_iniz;
    
% Cycle over the time steps
for n = 1 : NT-1 
    % 1st step : dt = dt_scheme/2 
    u_old = u(n);
    u_new = (u(n) + u(n+1)) / 2;
    f_mid = first_ode (NX, dx, x, dt*0.5, f_num(:,n), t(n), t(n) + dt*0.5, u_old, u_new);    
        
    % 2nd step: integral part (trapezoidal rule)  
    f_star = second_ode (dt, f_mid, G, u, dx);
       
    % 3rd step : Chang Cooper with dt = dt_scheme/2
    u_old = u_new;
    u_new = u(n+1);
    f_new = first_ode (NX, dx, x, dt*0.5, f_star, t(n) + dt*0.5, t(n+1), u_old, u_new); 
        
    % The numerical solution at time (n+1) is stored
    f_num(:,n+1) = f_new';   

end

end